%    This file is part of the psPack simulation package from the
%    doctoral dissertation "A DAE Formulation for Multi-Zone
%    Application to CVCP Engines" by Per Öberg. 
%    The dissertation is available at 
%    http://urn.kb.se/resolve?urn=urn:nbn:se:liu:diva-18120
%
%    Copyright (C) 2013  Per Öberg <per.oberg@liu.se>
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License along
%    with this program; if not, write to the Free Software Foundation, Inc.,
%    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

function [dXpT dXgc] = psDX_SingleZone_ConstV(XpT,V,Xgc,dm,dQ,TUp,XgcUp,dmUp,TDown,XgcDown,dmDown);
% function dX = psDX(XpVT,Xgc,dV,dm,dQ,extFlow);
%    XpT    = [p1, p2 , ... , T1, T2, ... ]'   (1+N)x1
%    Xgc     = [x_uf x_ua x_bf  x_ba]         NxK matrix
%    dm      = Mass flow between zones        N x N matrix 
%    dQ      = Heat transfer from resp zone   Nx1
%  External mass flow: There may only be flow to the first and last zone
%    TDown    = Temperture outside the control volume  Mx1
%    XgcDown  = Gas composition upstreams              MxK
%    dmDown   = Mass flow TO the control volume        Mx1
%      * Note the sign/direction convention
%    TUp    = Temperture outside the control volume  Mx1
%    XgcUp  = Gas composition upstreams              MxK
%    dmUp   = Mass flow TO the control volume        Mx1
%      * Note the sign/direction convention
% Output:
%    dXpT = 2*Nx1
%    dXgc  = NxK
%
%
%
% OVERVIEW OF dm :
%   _________________
%   I dm11 ... dm1n I  To zone 1
%   I dm21 ... dm2n I          2
%   I   .  .    .   I          .
%   I   .   .   .   I          .
%   I   .     . .   I          .
%   I dmn1 ... dmnn I          n
%   -----------------
%      1   ...   n
%   From zone
%
% This means:
%  * dm(i,j) > 0 => gas is entering zone i from zone j
%  * dm(i,j) < 0 => gas is leaving zone i for zone j 
%
%
% Author: Per Öberg (renewed implementation of old psPack with multiple
%                    authors)
%
% Changelog: 
%  2007-12-17 - Created


N = length(XpT)/2;
global psSimStruct
K = psSimStruct.psThermProp.K;

% Define index vecetors for the different states
% and use only these to avoid indexation mistakes.
pIdx = (1:N); 
tIdx = (1:N)+N;

% No data in X can be <0, physical limits for volume,
% pressure and temperature. This could generate a warning but since
% this will most probably only happen for calculations that will be
% discarded it does not seem necessary.

% Cylinder pressure is element 1 of input X
p = XpT(pIdx);  
T = XpT(tIdx);

if min(p) <= 0 || min(T ) <= 0 || min(V) <= 0
    dXpT = NaN*ones(size(XpT));
    dXgc = NaN*ones(size(Xgc));
    return
end
      

% This is where we get the thermodynamic properties
% We need:
% 1) For each zone everything is needed 
outputs = ones(1,9);
h    = zeros(N,1);
cp   = zeros(N,1);
cv   = zeros(N,1);
R    = zeros(N,1);
dRdp = zeros(N,1);
dRdT = zeros(N,1);
dudp = zeros(N,1);
dudX = zeros(N,K);
dRdX = zeros(N,K);

for ii = 1:N
    [h(ii) cp(ii) cv(ii) R(ii) dRdp(ii)...
     dRdT(ii) dudp(ii) dudX(ii,:) dRdX(ii,:)]=psThermProp(p(ii),T(ii),Xgc(ii,:),outputs);
end

% 2) for external zones only those who are from in-flows enthalpy is
% realy needed. For now, however, the implementation calculates 
% everything no matter if it's already been calulated before.
T_hat_Up   = (dmUp <= 0 )*T(1)+(dmUp > 0 ).*TUp;
T_hat_Down = (dmDown <= 0 )*T(end)+(dmDown > 0 ).*TDown;

Xgc_hat_Up = repmat((dmUp > 0),1,K).*XgcUp+...
    (dmUp <= 0)*Xgc(1,:);
Xgc_hat_Down = repmat((dmDown > 0),1,K).*XgcDown+...
    (dmDown <= 0)*Xgc(end,:);


outputs = zeros(1,9);
outputs(1) = 1;
h_hat_Up = psThermProp(p(1),T_hat_Up,Xgc_hat_Up,outputs);
h_hat_Down = psThermProp(p(end),T_hat_Down,Xgc_hat_Down,outputs);
    
% Calculate dXgc as soon as possible
mdXgc = zeros(N,K);
for ii = 1:N
    % Calculate m*dXgc for zone ii
    mdXgc(ii,:) = (dm(ii,:).*(dm(ii,:)>0))*(Xgc-repmat(Xgc(ii,:),N,1));
end
N1 = size(dmUp,1);
mdXgc(1,:)=mdXgc(1,:)+dot((Xgc_hat_Up-repmat(Xgc(1,:),N1,1)),repmat(dmUp.*(dmUp>0),1,K),1);
N2 = size(dmDown,1);
mdXgc(end,:)=mdXgc(end,:)+dot((Xgc_hat_Down-repmat(Xgc(end,:),N2,1)),repmat(dmDown.*(dmDown>0),1,K),1);
%dXgc=mdXgc/m;
dXgc=repmat(R.*T./(p.*V),1,K).*mdXgc;
%disp('Warning: There has to be a more efficient way!')
%if any(dXgc(:))
%    disp('Debug: in psDX.m We have flow')
%end


%dXgc = dXgc';
if ~isreal(sum(dXgc))
    disp('Error in psDX.m: Complex dXgc')
end


% For internal flows we have everything we need by now. We only need to
% structure it a bit.
%  h_hat_int (internal) is the enthalpy of the gas that enters or leaves
%  the zones internally
%  * dm(i,j) > 0 => gas is entering zone i from zone j => h_hat_int(i,j) = h(j)
%  * dm(i,j) < 0 => gas is leaving zone i for zone j => h_hat_int(i,j) = h(i)
%  * dm(i,i) = 0 sinze we have no mass tranfer to ourselfes.
%  * h(i,i) = hi(i) (But could be set to zero).
%  * hij is thus symmetric, but dm is antisymmetric.
%
% Remember that h is a vector with entalpies.
% The right elements of h shall now be placed at the rigt positions in 
% h_hat_int. We can use the fact that depending on the sign of dm(i,j) we shold
% pick either [h ... h](i,j) or [h' ; ... ; h'](i,j). Since h_hat_int is
% symmetric we can take those elements of [h ... h] where dm is
% positive and put them in h_hat_int. We can then transpose h_hat_int and
% add it to itself. After that only those elements in h_hat_int that
% corresponds to  dm(i,j)=0 is missing (e.g. the diagonal of h_hat_int.)
h_hat_int = (sign(dm) == 1 ).*(repmat(h',N,1));
h_hat_int = h_hat_int + h_hat_int';%+diag(hi) (Removed to save comuptation time since it is
                 %unused)
% Note: h_hat_int(i,j)=0 for all elements that correpsonds to dm(i,j)=0
% after this operation. However, as will be seen later 
% h_hat_int for elements with dm(i,j)=0 does not enter the equation at all.



% Build the matrix components and then finally the A-matrix. Note that it's
% the same for both "well stirred mixer" and "Well stirred reactor" since
% dRdp=dRdT=dudp=0 in the well stirred mixer case. The matrix is shifted
% for easier implementation
a = V-p.*V./R.*dRdp;
b =-p.*V./T-p.*V./R.*dRdT;
c = p.*V./R./T.*dudp;
d = p.*V./R./T.*cv;

Adet = a.*d-b.*c;
Ainv = [diag(d./Adet) diag(-b./Adet) ;...
        diag(-c./Adet) diag(a./Adet)];
% h = u + pv = u + RT => u = h - RT
B1      = R.*T.*sum(dm,2)+...
          dot(dRdX,mdXgc,2).*T;
B1(1)   = B1(1)+R(1)*T(1)*sum(dmUp);
B1(end) = B1(end)+R(end)*T(end)*sum(dmDown);

B2 = -dQ-dot(dudX,mdXgc,2)+...
     sum((h_hat_int-repmat(h-R.*T,1,N)).*dm,2);

B2(1)   = B2(1)   +(h_hat_Up-h(1)+R(1)*T(1))'*dmUp;
B2(end) = B2(end) +(h_hat_Down-h(end)+R(end)*T(end))'*dmDown;

B = [B1;B2];

dXpT= Ainv*B;

