%    This file is part of the psPack simulation package from the
%    doctoral dissertation "A DAE Formulation for Multi-Zone
%    Application to CVCP Engines" by Per Öberg. 
%    The dissertation is available at 
%    http://urn.kb.se/resolve?urn=urn:nbn:se:liu:diva-18120
%
%    Copyright (C) 2013  Per Öberg <per.oberg@liu.se>
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License along
%    with this program; if not, write to the Free Software Foundation, Inc.,
%    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

function X0 = initCombustion(theta,Xend)
% Initializes combustion phase. b = burned zone, u = unburned zone

% Xend = [p,Vu,Tu,-Xc-] from gasexchange


global psSimStruct

K = psSimStruct.psThermProp.K;

if psSimStruct.psSimPar.giveWarnings
  disp('---')
  disp('Initializing combustion phase')
  %disp('-- press enter --')
  %pause
end

%%% Show progressbar
%showProgress();

% -------------------------------
% Get some parameters
% -------------------------------
initBurnedVolFrac = psSimStruct.psSimPar.initBurnedVolFrac;
deltaT_ad = psSimStruct.psSimPar.deltaT_ad;
%AFs = psSimStruct.psThermProp.AFs;
minVFrac = psSimStruct.psSimPar.minVFrac;
%-----------------------------
% Checking for fresh start
%-----------------------------
if nargin < 2
    if psSimStruct.psSimPar.giveWarnings
    disp('Detected fresh initialization. Reading values from workspace.')
    end
    if minVFrac < psVibe(theta)
        xbInit = psVibe(theta);
    else
        xbInit =minVFrac;
    end
    disp('psInitCombustion: Fix the initial values for the combustion zone')

    % Start out with a temperetaure of 300K and a volumetric efficiency of about
    % 0.9 with a residual mass fraction of 5% and assume Ideal otto cycle up
    % until the current crank angle. Also, assume phi = 1.
    
    %pV^gamma= const
    p0 = psSimStruct.psModel.pIC_IM*0.9;
    T0 = 300;
    p1=p0*(psVolume(pi)/psVolume(theta))^1.35;

    mR = p0*psVolume(pi)/T0;
    
    T1 = T0*(psVolume(pi)/psVolume(theta))^0.35;
    % du/dT = cv (cv const) => qHV*/(1+AFs) = cv*delta_T
    [h cp cv ] = psThermProp(p1,2000,psCalcXgc(1,1,1,0,1,0),ones(1,9));    
    if isfield(psSimStruct.psThermProp,'qLHVEst')
        delta_T = 0.95*psSimStruct.psThermProp.qLHVEst/(1+psSimStruct.psThermProp.AFs)/cv;
    else
        delta_T = 0.95*44e6/(1+psSimStruct.psThermProp.AFs)/cv;
    end
    
    T2 = T1+delta_T;
    
    
    Tu = T1;
    Tb = T2;

    Ttot = Tu*(1-xbInit)+xbInit*Tb;
   
    p2 = mR*Ttot/psVolume(theta);
    

    Xc_b = psCalcXgc(1,1,1,0,1,0);%psSimStruct.psModel.XcIC_b_Cb;
    Xc_u = psCalcXgc(1,0,1,0.05,1,0.95);%psSimStruct.psModel.XcIC_u_Cb;

    Vb_prop=xbInit*Tb;
    Vu_prop=(1-xbInit)*Tu;
  
    propFac = (Vb_prop+Vu_prop)/psVolume(theta);
    
    Vb = Vb_prop/propFac;
    Vu = Vu_prop/propFac;
    
    X0   = [p2 Vb Tb Vu Tu Xc_b  Xc_u];
else
    p    = Xend(1);
    T    = Xend(3);
    Xc   = Xend(3+(1:K))';
    [h_u cp_u cv_u R_u] = psThermProp(p,T,Xc,[1 1 1 1 0 0 0 0 0]);
    Xc_u = Xc;
    Xc_b = psCalcReactXgc(Xc_u);
    
    % -------------------------------
    % Calculate temperature and volume
    % -------------------------------
    % This is an approximation of the first little burned fraction
    V  = psVolume(theta); % Fix for driftoff from psVolume
    Vb = initBurnedVolFrac*V;
    Vu = (1-initBurnedVolFrac)*V;
    Tu = T;
    % An initial guess
    Tb = Tu+deltaT_ad; % *(1-xRes_u)*phi_u;
    deltaT = Tb;
    %disp('In psInitCombustion: WARNING deltaT_ad is a really crude approximation')
    lpCount = 0;
    while abs(deltaT) > 1e-6*Tb && lpCount < 100
        [h_b cp_b cv_b R_b] = psThermProp(p,Tb,Xc_b,[1 1 1 1 0 0 0 0 0]);
        deltaT  = -(h_b-h_u)/cp_b;
        Tb = deltaT+Tb;
        lpCount = lpCount+1;
    end    
    if lpCount >= 100
        disp('Warning: Loop Counter in InitCombustion >= 10')
    end
    X0 = [p; Vb; Tb; Vu; Tu; Xc_b'; Xc_u'];
end



