%    This file is part of the psPack simulation package from the
%    doctoral dissertation "A DAE Formulation for Multi-Zone
%    Application to CVCP Engines" by Per Öberg. 
%    The dissertation is available at 
%    http://urn.kb.se/resolve?urn=urn:nbn:se:liu:diva-18120
%
%    Copyright (C) 2013  Per Öberg <per.oberg@liu.se>
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License along
%    with this program; if not, write to the Free Software Foundation, Inc.,
%    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

function ps2ZoneCylinder(block)
setup(block);
%endfunction

%% Function: setup ===================================================
%% Abstract:
%%   Set up the S-function block's basic characteristics such as:
%%   - Input ports
%%   - Output ports
%%   - Dialog parameters
%%   - Options
%% 
%%   Required         : Yes
%%   C-Mex counterpart: mdlInitializeSizes
%%
function setup(block)
  %disp(sprintf('Running ps2ZoneCylinder:Setup at time: %d',NaN))
  % Register number of ports
  global psSimStruct
  K = psSimStruct.psThermProp.K;
      
  block.NumInputPorts  = 2;
  block.NumOutputPorts = 1;

  
  % In signals: 
  % 1) theta
  % 4) Ts (T_Wall)
  % Utsignaler:
  % 1) DiagOut (Cont states Disc states and dQ)

  % Cont States are:
  %  X = [p,Vb,Tb,Vu,Tu,-Xgc_u-,-Xgc_b-]
  % Disc States are:
  %   pivc,Vivc,Tivc
  
  
  % Update sker efter output i varje steg. Dvs saker händer i följande ordning:
  % 1) Setup
  % 2) SetInPortFrameData
  % 3) DoPostPropSetup
  % 4) Start (time exists from hereon)
  % 5) Init 
  % 6) Output
  % 7) Update
  % 8) Repeat 6-7 and possibly 5-7
  % Start kan alltså anvädas första ggn modellen startas för att
  % avgöra vars vi är  medans initialize kan köras varje gång vi kommer
  % tillbaks från ett disablat tillstånd.

  % None of this dynamic inport stuff
  block.SetPreCompPortInfoToDefaults;
     
  %%%%%%%%%%%%%%%%
  % Input ports
  %%%%%%%%%%%%%%%%
  % Data type
  block.InputPort(1).DatatypeID  = 0;  % double
  block.InputPort(2).DatatypeID  = 0;  % double
  % block.InputPort(3).DatatypeID  = 0;  % double

  % No complex signals
  block.InputPort(1).Complexity  = 'Real';
  block.InputPort(2).Complexity  = 'Real';
  % block.InputPort(3).Complexity  = 'Real';

  % Dimensions are fixed
  block.InputPort(1).Dimensions  = 1;
  block.InputPort(2).Dimensions  = 1;
  % block.InputPort(3).Dimensions  = 1;

  % No direct feedthrough
  block.InputPort(1).DirectFeedthrough = false;
  block.InputPort(2).DirectFeedthrough = false;
  % block.InputPort(3).DirectFeedthrough = false;

  % Never accept frame data... 
  block.InputPort(1).SamplingMode  = 0;
  block.InputPort(2).SamplingMode  = 0;
  % block.InputPort(3).SamplingMode  = 0;

  %%%%%%%%%%%%%%%%
  % Output ports
  %%%%%%%%%%%%%%%%  
  % Data type
  block.OutputPort(1).DatatypeID  = 0; % double

  % No complex signals
  block.OutputPort(1).Complexity  = 'Real';
  
  % Dimensions are fixed
  block.OutputPort(1).Dimensions  = 5+K*2+3+1;

  % Never accept frame data... 
  block.OutputPort(1).SamplingMode  = 0;
  
  % Register parameters
  block.NumDialogPrms     = 2;
  %  block.DialogPrmsTunable = {'Tunable','Nontunable','SimOnlyTunable'};

  % Register sample times
  %  [0 offset]            : Continuous sample time
  %  [positive_num offset] : Discrete sample time
  %
  %  [-1, 0]               : Port-based sample time
  %  [-2, 0]               : Variable sample time
  block.SampleTimes = [0 0];
  
  % Cont States are:
  %  X = [p,Vb,Tb,Vu,Tu,-Xc_u-,-Xc_b-]
  % Disc States are:
  %   pivc,Vivc,Tivc
  block.NumContStates = 5+2*K;
  
  %% -----------------------------------------------------------------
  %% Options
  %% -----------------------------------------------------------------
  % Specify if Accelerator should use TLC or call back into 
  % M-file
  block.SetAccelRunOnTLC(false);
  
  %% -----------------------------------------------------------------
  %% Register methods called during update diagram/compilation
  %% -----------------------------------------------------------------
  
  %% 
  %% CheckParameters:
  %%   Functionality    : Called in order to allow validation of
  %%                      block's dialog parameters. User is 
  %%                      responsible for calling this method
  %%                      explicitly at the start of the setup method
  %%   C-Mex counterpart: mdlCheckParameters
  %%
  block.RegBlockMethod('CheckParameters', @CheckPrms);

  %%
  %% PostPropagationSetup:
  %%   Functionality    : Setup work areas and state variables. Can
  %%                      also register run-time methods here
  %%   C-Mex counterpart: mdlSetWorkWidths
  %%
  block.RegBlockMethod('PostPropagationSetup', @DoPostPropSetup);

  %% -----------------------------------------------------------------
  %% Register methods called at run-time
  %% -----------------------------------------------------------------
  
  %% 
  %% ProcessParameters:
  %%   Functionality    : Called in order to allow update of run-time
  %%                      parameters
  %%   C-Mex counterpart: mdlProcessParameters
  %%  
  block.RegBlockMethod('ProcessParameters', @ProcessPrms);

  %% 
  %% InitializeConditions:
  %%   Functionality    : Called in order to initalize state and work
  %%                      area values
  %%   C-Mex counterpart: mdlInitializeConditions
  %% 
  block.RegBlockMethod('InitializeConditions', @InitializeConditions);
  
  %% 
  %% Start:
  %%   Functionality    : Called in order to initalize state and work
  %%                      area values
  %%   C-Mex counterpart: mdlStart
  %%
  block.RegBlockMethod('Start', @Start);

  %% 
  %% Outputs:
  %%   Functionality    : Called to generate block outputs in
  %%                      simulation step
  %%   C-Mex counterpart: mdlOutputs
  %%
  block.RegBlockMethod('Outputs', @Outputs);

  %% 
  %% Update:
  %%   Functionality    : Called to update discrete states
  %%                      during simulation step
  %%   C-Mex counterpart: mdlUpdate
  %%
  block.RegBlockMethod('Update', @Update);

  %% 
  %% Derivatives:
  %%   Functionality    : Called to update derivatives of
  %%                      continuous states during simulation step
  %%   C-Mex counterpart: mdlDerivatives
  %%
  block.RegBlockMethod('Derivatives', @Derivatives);
  

  %% 
  %% ZeroCrossings:
  %%   Functionality    : For S-functions with continuous sample
  %%                      time, if input signals have discontinuities,
  %%                      this method is called to detect 
  %%                      discontinuities
  %%   C-Mex counterpart: mdlZeroCrossings
  %%
  block.RegBlockMethod('ZeroCrossings', @ZeroCrosssings);
  
  %% 
  %% SimStatusChange:
  %%   Functionality    : Called when simulation goes to pause mode
  %%                      or continnues from pause mode
  %%   C-Mex counterpart: mdlSimStatusChange
  %%
  block.RegBlockMethod('SimStatusChange', @SimStatusChange);
  
  %% 
  %% Terminate:
  %%   Functionality    : Called at the end of simulation for cleanup
  %%   C-Mex counterpart: mdlTerminate
  %%
  block.RegBlockMethod('Terminate', @Terminate);

  %% -----------------------------------------------------------------
  %% Register methods called during code generation
  %% -----------------------------------------------------------------
  
  %%
  %% WriteRTW:
  %%   Functionality    : Write specific information to RTW file
  %%   C-Mex counterpart: mdlRTW
  %%
  block.RegBlockMethod('WriteRTW', @WriteRTW);

%endfunction

%% -------------------------------------------------------------------
%% The local functions below are provided for illustrative purposes
%% to show how you may implement the various block methods listed
%% above.
%% -------------------------------------------------------------------

function CheckPrms(block)
    %disp(sprintf('Running ps2ZoneCylinder:CheckPrms at time: %d',NaN))
%  a = block.DialogPrm(1).Data;
%  if ~strcmp(class(a), 'double')
%    error('Invalid parameter');
%  end
  
%endfunction

function ProcessPrms(block)
  %disp(sprintf('Running ps2ZoneCylinder:ProcessPrms at time: %d',block.currentTime))
  block.AutoUpdateRuntimePrms;
 
%endfunction
  
function DoPostPropSetup(block)
    %disp(sprintf('Running ps2ZoneCylinder:DoPostPropSetup at time: %d',NaN))
    block.NumDworks = 2;

    % pivc,Vivc,Tivc
    block.Dwork(1).Name            = 'staticStates';
    block.Dwork(1).Dimensions      = 3;
    block.Dwork(1).DatatypeID      = 0;      % double
    block.Dwork(1).Complexity      = 'Real'; % real
    block.Dwork(1).UsedAsDiscState = true;

    block.Dwork(2).Name            = 'mdlStartTime';
    block.Dwork(2).Dimensions      = 1;
    block.Dwork(2).DatatypeID      = 0;      % double
    block.Dwork(2).Complexity      = 'Real'; % real
    block.Dwork(2).UsedAsDiscState = true;
    
    %% Register all tunable parameters as runtime parameters.
    block.AutoRegRuntimePrms;

%endfunction

function InitializeConditions(block)
  global psSimStruct
  global setBurnedTemp
  %CylNbr = block.DialogPrm(2).Data;
  
  %disp(sprintf('Running ps2ZoneCylinder:InitConditions at time: %d',...
  %             block.currentTime));
  caOffset = block.DialogPrm(1).Data;
  currentCa = block.currentTime*psSimStruct.psSimPar.w+caOffset;
  
  initStates = psInitCombustion(currentCa);
  
  if ~isempty(setBurnedTemp)
      disp(sprintf('Setting Tb=%d',setBurnedTemp))
      initStates(3) = setBurnedTemp;
  end
  % X0=[p,Vb,Tb,Vu,Tu,-Xgc_b-,-Xgc_u-];
  block.ContStates.Data = initStates;
  block.Dwork(1).Data = initStates([1,4,5]);
  % pivc,Vivc,Tivc
%endfunction

function Start(block)
  global psSharedSimData
  CylNbr = block.DialogPrm(2).Data;
  %disp(sprintf('Running ps2ZoneCylinder:Start at time: %d',block.currentTime))
  % Save start time in Dwork(2)  to be able to use it for initializing later
  block.Dwork(2).Data = block.currentTime;
  % Make sure that the block is saved in the shared simdata
  psSharedSimData.CombBlock(CylNbr) = block;

%endfunction

function WriteRTW(block)
   block.WriteRTWParam('matrix', 'M',    [1 2; 3 4]);
   block.WriteRTWParam('string', 'Mode', 'Auto');
   
%endfunction

function Outputs(block)
%  %disp(sprintf('Running ps2ZoneCylinder:Outputs at time: %d',...
%                block.currentTime))

  theta    = block.InputPort(1).Data(1);
  Ts       = block.InputPort(2).Data(1);
  pivc     = block.Dwork(1).Data(1);
  Vivc     = block.Dwork(1).Data(2);
  Tivc     = block.Dwork(1).Data(3);
  
  global psSimStruct
  % Get some parameters
  K = psSimStruct.psThermProp.K;
  w = psSimStruct.psSimPar.w;
  % Gamma for heat transfer calculation
  gamma	= psSimStruct.psThermProp.gammaApprox;  
  % Slaglängd
  S = psSimStruct.psGeometry.S;
  
  % Pack the states into X
  %  X = [p,Vb,Tb,Vu,Tu]';
  X  = block.ContStates.Data';
  p  = X(1);
  Tb = X(3);
  Tu = X(5);
  XpVT  = X(1:5)';
  Xgc_b = block.ContStates.Data(5+(1:K))';
  Xgc_u = block.ContStates.Data(5+K+(1:K))';
  Xgc = [Xgc_b ; Xgc_u];
  
  V = psVolume(theta);
  A   = psArea(theta);
  xB  = psVibe(theta); % If this version of xB is chosen it will be a lot
                        % less heat transfer for the freshly initialized
                        % burned zone during the compression. This will
                        % speed up the simulation considerably. (Most
                        % important in the update part)
  %xB = m_b/m_tot;
  
  pm  = pivc*(Vivc/V)^gamma;
  mps = S*w/pi;
  
  if Tb < 100
      Tb = 100;
  end
  if Tb > 4000
      Tb = 4000;
  end
  
  if pm > p
       pm = p;
  end

  
  dQ(1,1) = psDQ(pm,p,Tb,V,A*abs(xB),    mps,pivc,Tivc,Vivc,Ts);
  dQ(2,1) = psDQ(pm,p,Tu,V,A*abs(1-xB),mps,pivc,Tivc,Vivc,Ts);
  if ~isreal(sum(dQ))
      %disp('dQ un-Real')
      % This means that the there is something wrong. Most likely it's
      % the tempearature that is negative, or the motored pressure that
      % is larger than the real pressure. This occurs sometimes for very
      % small zones when simulink tries to take a very large
      % step. Therefore  dQ could be set to zeero to NaN to 
      % avoid this problem.
      dQ = NaN;      
  end

  

  % Diag port
  block.OutputPort(1).Data = ...
      [block.ContStates.Data ;...
       block.Dwork(1).Data ;...
       sum(dQ)]; 
  
%endfunction

function Update(block)
%endfunction

function Derivatives(block)
%   disp(sprintf('Running ps2ZoneCylinder:Derivatives at time: %d',...
%                 block.currentTime))

  theta    = block.InputPort(1).Data(1);
  Ts       = block.InputPort(2).Data(1);

  global psSimStruct
  K = psSimStruct.psThermProp.K;
  % Get some parameters
  w = psSimStruct.psSimPar.w;
  AFs = psSimStruct.psThermProp.AFs;

  % Gamma for heat transfer calculation
  gamma	= psSimStruct.psThermProp.gammaApprox;  

  S = psSimStruct.psGeometry.S;
  
  % Pack the states into X
  %  X = [p,Vb,Tb,Vu,Tu,-Xc_u-,-Xc_b-]';
  X     = block.ContStates.Data';
  p     = X(1);
  Vb    = X(2);
  Tb    = X(3);
  Vu    = X(4);
  Tu    = X(5);
  XpVT  = X(1:5)';
  Xgc_b = block.ContStates.Data(5+(1:K))';
  Xgc_u = block.ContStates.Data(5+K+(1:K))';
  pivc  = block.Dwork(1).Data(1);
  Vivc  = block.Dwork(1).Data(2);
  Tivc  = block.Dwork(1).Data(3);

  Xgc = [Xgc_b ; Xgc_u];

  % Calculate thermochemical properties: h, cp, R, dRdp, dRdT
  % All zones and all parts have the same phi so just set them rigt
  % here. 
  [dummy1 dumy2 dummy3 R] = psThermProp(p,[Tb; Tu],Xgc,[0 0 0 1 0 0 0 0 0]);   
  % Define massflow to and from the zones   
  % dm(i,j) is the massflow from zone j to zone i.
  % Massflows entering zone i have a positive sign.

%Note:
% Anledningen till att man vill ha två zoner under förbränningsfasen är
% att man vill kunna beräkna NOx
  m_b = p*Vb/R(1)/Tb;
  m_u = p*Vu/R(2)/Tu;
  m_tot = m_b+m_u;
  xB    = m_b/m_tot;
  %  if ~isfinite(m_tot)
%    out = realmax*ones(2*zones+2,1);
%    return
%  end
  
  dm = zeros(2,2);

  if Vu > (Vb+Vu)*psSimStruct.psSimPar.minVFrac
      dm(1,2) = m_tot*w*psDVibe(theta,xB);
      dm(2,1) = -dm(1,2);
  end

  % Define heat transfer into the zones   
  %V   = psVolume(theta);
%  Vb
%  Vu
%
  V = psVolume(theta);
%  V   = Vb + Vu;
  A   = psArea(theta);
  xB  = psVibe(theta); % If this version of xB is chosen it will be a lot
                        % less heat transfer for the freshly initialized
                        % burned zone during the compression if the burned
                        % zone is initialized to a non-zero-size. This will
                        % speed up the simulation considerably.
  %xB = m_b/m_tot;
  pm  = pivc*(Vivc/V)^gamma;
  mps = S*w/pi;

  if Tb < 100
      Tb = 100;
  end
  if Tb > 4000
      Tb = 4000;
  end
     
  if pm > p
       pm = p;
  end
  
  dQ(1,1) =psDQ(pm,p,Tb,V,A*abs(xB),    mps,pivc,Tivc,Vivc,Ts);
  dQ(2,1) =psDQ(pm,p,Tu,V,A*abs(1-xB),mps,pivc,Tivc,Vivc,Ts);
  % Define total derivative of total cylinder Volume
  dV = w*psDVolume(theta);
  
  Xgc_to = psCalcReactXgc(Xgc);

  [dXpVT dXgc]=  psDX(XpVT,Xgc,Xgc_to,dV,dm,dQ,zeros(2,2),zeros(2,K,2),zeros(2,K,2),zeros(2,2));  
  if ~isreal(sum(dXpVT))|| ~isreal(sum(dXgc))
      % This means that the there is something wrong. Most likely it's
      % the tempearature that is negative, or the motored pressure that
      % is larger than the real pressure. This occurs sometimes for very
      % small zones when simulink tries to take a very large
      % step. Therefore  The values could be set to NaN so that Simulink
      % detects the error and takes a shorter step
      dXpVT = NaN(size(dXpVT));
      dXgc  = NaN(size(dXgc));
  end

  block.Derivatives.Data = [dXpVT(:) ;dXgc(1,:)';dXgc(2,:)'];

%endfunction

function ZeroCrosssings(block)
  % Eventuellt lägga till lite kollar för zero-cross hjälp.
  % Tillstånden är:
  % - p,Vbg,Tbg,Vfm,Tfm,Tav,phi_bg,phiRes_bg,xRes_bg,phi_fm,phiRes_fm,xRes_fm
  % och det som kan hända med dem är att Volymerna kan närma sig gränsen
  % för vad som är acceptabel storlek. De viktigaste eventsen är redan
  % omhändertagna men även events som ändring av riktning på flöden mm
  % är av intresse. Ett bra sätt är att kolla efter if-satser i
  % derivata-uträkningen och sätta zero-cross-detection på signalerna i
  % omgivningen som har med signalen att göra, om det är viktigt vill säga.
%endfunction

function SimStatusChange(block, s)
  
  if s == 0
    disp('Pause has been called');
  elseif s == 1
    disp('Continue has been called');
  end
  
%endfunction
    
function Terminate(block)
%endfunction
 
